//=======================================================================
//
//  Language.java
//
//  Description: Language Engine
//
// An Engine to generate random words that resemble a particular 
// language, based on statistical information about that language.
//
// Based on the Wordgen source code in Euphoria by Jeff Zeitlin
// 
// To do ? : Change to use .properties or XML language source files instead of 
//           non-standard Windows format ?
//         : Optimise read and generation 
//         :
// 
//=======================================================================
//  Copyright (c) Frank G. Pitt, 1 June 2000
//
//  Permission is hereby granted to use or modify this code for 
//  non-commercial purposes as long as the above copyright is 
//  retained or reproduced in any derivative work. 
//
//=======================================================================
//                         Maintenance Log
//------------|-------------|--------------------------------------------
// dd/mm/yy   | Author      |                 Description
//------------|-------------|--------------------------------------------
// 10/08/2000 | Frank Pitt  | Created
//------------|-------------|--------------------------------------------
// ??/??/2000 |             | 
//------------|-------------|--------------------------------------------
//   Record your modifications !
//=======================================================================
// package gen.mundens.rpg.language

import java.io.* ;
import java.util.* ;

class Language
{
  public Language( String name )
  {
    // initialize Randomm only oncce per run !
    if ( this.rand == null )
      rand = new Random() ;
    this.name = name ; 
    loadTables() ;
  }

  private void loadTables() 
  {
    String languageFilename = this.name + ".lng" ;
    String line ;
   
    try 
    {
      BufferedReader br = new BufferedReader( new FileReader( languageFilename ) ) ;
      line = br.readLine() ;

      do
      {
       processLine( line ) ;  
       line = br.readLine() ;
      }
      while (line != null ) ;
    }
    catch ( IOException ioe )
    {
       System.err.println(" Some file error happened loading the language table specified" );
       System.exit(0);
    }
  }

  private void processLine( String line )
  {
    Hashtable table ;
    String key ;
    String val ;
 
    if ( ( line.length()  == 0   ) ||     // blank line, ignore
         ( line.charAt(0) == ';' ) )    // comment, ignore
        return ;
    
    if ( line.charAt(0) == '['  )    // section start
    {
      tableName = (line.substring( 1 , line.length() - 1 )).toUpperCase() ;
      table = new Hashtable() ;
      tables.put (  tableName, table ) ;
    }
    else                        // presumed key=value line
    {
       StringTokenizer st = new StringTokenizer( line, "=" ) ;
       key = st.nextToken() ;
       val = st.nextToken() ;

        if ( ( tableName.equals("SYLLABLES") ) || 
             ( tableName.equals("ASSIMILATE") )
           )
        {
          ((Hashtable) tables.get( tableName )).put( key, val ) ;
        }
        else
        {
          try 
          {
           int value= Integer.parseInt( val ) ;
           
           // copying Jeff's Euphoria code, we repeat the key the same number 
           // of times as it's relative frequency, making data structures larger,
           // but randomization much simpler 
           while ( value >0 )   
           { 
             ((Hashtable) tables.get( tableName )).put( key, new Integer(value--) ) ;
           }   

          }
          catch ( NumberFormatException nfe )
          {
            System.err.println( "Non-numeric frequency of occurrence.\n") ;
            System.exit(1) ;
          }
        }
     }
  }

  public String makeWord()
  { 
     return assimilate( doMakeWord( getNumberOfSyllables() ) );
  }

  private int getNumberOfSyllables()
  {
    Hashtable lengthTable = (Hashtable) tables.get("LENGTH") ;

    if ( ( lengthTable = null ) ||  ( lengthTable.size() = 0 ) )
    {
        syllables = rand.nextInt( 5 ) + 1 ;
    }
    else
    {
      numSyllables = (String) initialTable.keySet().toArray()[ rand.nextInt( initialTable.size() ) ] ;
      try 
      {
        syllables = Integer.parseInt( numSyllables ) ;
      }
      catch ( NumberFormatException nfe )
      {
        System.err.println( "Non-numeric frequency of occurrence.\n") ;
        System.exit(1) ;
      }
    }
    return syllables ; 
  }

  private String doMakeWord( int numSyllables )
  { 
    String word = new String() ;

    Hashtable syllablesTable = (Hashtable) tables.get("SYLLABLES") ;
    String syllableType = "FIRST" ;

    for ( int i = 0 ; i < numSyllables ; i++ )
    {

      String    sourceTableName = (String) syllablesTable.get( syllableType ) ;
      Hashtable sourceTable     = (Hashtable) tables.get( sourceTableName ) ;

      if ( sourceTable.size() == 0) 
            return word ;

      String syltype =  (String) sourceTable.keySet().toArray()[ rand.nextInt( sourceTable.size() ) ] ;


      if ( syltype.charAt(0) == 'C' )
      {
        Hashtable initialTable = (Hashtable) tables.get("INITIAL") ;
        word = word + (String) initialTable.keySet().toArray()[ rand.nextInt( initialTable.size() ) ] ;
      }

      Hashtable vowelTable = (Hashtable) tables.get("VOWEL") ;
      word = word + (String) vowelTable.keySet().toArray()[ rand.nextInt( vowelTable.size() ) ] ;

      if ( syltype.charAt( syltype.length() - 1) == 'C' )
      {
        Hashtable finalTable = (Hashtable) tables.get("FINAL") ;
        word = word + (String) finalTable.keySet().toArray()[ rand.nextInt( finalTable.size() ) ] ;
      }
    }
    return word ;
  }

  private String assimilate( String word )
  {
    Hashtable assimilations = (Hashtable) tables.get("ASSIMILATE") ;

    if ( assimilations.size() > 0) 
    {
        Iterator iter  = assimilations.keySet().iterator() ;
 
        while ( iter.hasNext() )
        {
            String key = (String) iter.next();

            int pos = word.indexOf( key ) ;  

            while ( pos != -1 ) 
            {
               word =  word.substring( 0, pos) 
                     + (String) assimilations.get( key )
                     + word.substring( pos + key.length() );

               pos = word.indexOf( key, pos + ((String) assimilations.get( key )).length() ) ;  

            }
        }
    }

    return word ;
  }

 
  // test that we read it correctly by writing it all out again
  public void write() 
  {
    Iterator iter = tables.keySet().iterator() ;

    while ( iter.hasNext() )
    {
      String key = (String) iter.next() ;
      System.out.println( "" ) ;
      System.out.println( "["  + key + "]") ;
      
      Hashtable table = (Hashtable) tables.get( key ) ; 
      
      Iterator iter2 = table.keySet().iterator() ;

      while ( iter2.hasNext() )
      {
        String key2 = (String) iter2.next() ;
        System.out.println( key2 + "=" + table.get( key2  ) ) ;
      } 
    }
  }

  // Main allows standalone testing of language engine 
  public static void main( String args[] )
  {
     Language language = new Language( "Dutch" ) ;
     System.out.println(language.makeWord()) ;
     System.out.println(language.makeWord()) ;
     System.out.println(language.makeWord()) ;
     System.out.println(language.makeWord()) ;

     language = new Language( "Maori" ) ;
     System.out.println(language.makeWord()) ;
     System.out.println(language.makeWord()) ;
     System.out.println(language.makeWord()) ;
     System.out.println(language.makeWord()) ;
  }

  private String name ;
  private Hashtable tables = new Hashtable();  

  // used to store table name between calls to processLine()
  private String tableName ;

  private static Random rand  ;


}
